from typing import List, Optional, Tuple

import numpy as np

from ..distributions import Debye, ColeCole, ColeDavidson, KWW, HavriliakNegami
from ..utils.constants import epsilon0


class _AbstractBDS:
    name = 'Abstract'
    type = 'Dielectric Spectroscopy'
    equation = ''
    params = [r'\Delta\epsilon', r'\tau_{0}']
    bounds = [(0, None), (0, None)]
    susceptibility = None
    iscomplex = True

    @classmethod
    def func(cls, x, *args, **kwargs):
        # args[0] : Delta epsilon
        # args[1:] : every other parameter
        chi = args[0] * cls.susceptibility(2*np.pi*x, *args[1:])

        return chi


class DebyeBDS(_AbstractBDS):
    name = 'Debye'
    equation = r'\Delta\epsilon / (1-i\omega\tau)'
    susceptibility = Debye.susceptibility


class ColeColeBDS(_AbstractBDS):
    name = 'Cole-Cole'
    equation = r'\Delta\epsilon / (1-i\omega\tau)^{\alpha}'
    params = _AbstractBDS.params + [r'\alpha']
    bounds = _AbstractBDS.bounds + [(0, 1)]
    susceptibility = ColeCole.susceptibility


class ColeDavidsonBDS(_AbstractBDS):
    name = 'Cole-Davidson'
    equation = r'\Delta\epsilon / [1-(i\omega\tau)^{\gamma})'
    params = _AbstractBDS.params + [r'\gamma']
    bounds = _AbstractBDS.bounds + [(0, 1)]
    susceptibility = ColeDavidson.susceptibility


class HavriliakNegamiBDS(_AbstractBDS):
    name = 'Havriliak-Negami'
    equation = r'\Delta\epsilon / [1-(i\omega\tau)^{\gamma}]^{\alpha}'
    params = _AbstractBDS.params + [r'\alpha', r'\gamma']
    bounds = _AbstractBDS.bounds + [(0, 1), (0, 1)]
    susceptibility = HavriliakNegami.susceptibility


class KWWBDS(_AbstractBDS):
    name = 'KWW'
    params = _AbstractBDS.params + [r'\beta']
    bounds = _AbstractBDS.bounds + [(0, 1)]
    susceptibility = KWW.susceptibility


class EpsInfty:
    name = 'Epsilon Infinity'
    type = 'Dielectric Spectroscopy'
    equation = r'\epsilon_{\infty}'
    params = [r'\epsilon_{\infty}']
    bounds = [(None, None)]
    iscomplex = True

    @staticmethod
    def func(x, eps):
        ret_val = np.zeros(x.shape, dtype=complex)
        ret_val += eps

        return ret_val


class PowerLawBDS:
    name = 'Power Law'
    type = 'Dielectric Spectroscopy'
    equation = r'A / (i\omega)^{n}'
    params = ['A', 'n']
    bounds = [(None, None), (None, None)]
    iscomplex = True

    @staticmethod
    def func(x, a, n):
        return a / (1j*x)**n


class DCCondBDS:
    name = 'DC cond.'
    type = 'Dielectric Spectroscopy'
    equation = r'i\sigma_{dc}/\epsilon_{0}\omega'
    params = [r'\sigma_{dc}']
    bounds = [(0, None)]
    iscomplex = True

    @staticmethod
    def func(x, sigma):
        ret_val = np.zeros(x.shape, dtype=complex)
        ret_val += 1j * sigma / x / epsilon0

        return ret_val


class HavriliakNegamiDerivative:
    name = 'Derivative HN'
    type = 'Dielectric Spectroscopy'
    params = [r'\Delta\epsilon', r'\tau', r'\alpha', r'\gamma']
    bounds = [(0, None), (0, None), (0, 1), (0, 1)]

    @staticmethod
    def func(x, eps, tau, a, g):
        omtau = 2*np.pi*x * tau
        theta = np.arctan2(np.sin(np.pi*a/2.), omtau**(-a) + np.cos(np.pi*a/2.))

        numer = a*g * omtau**a * np.cos(np.pi*a/2. - (1+g)*theta)
        denom = (1 + 2*omtau**a * np.cos(np.pi*a/2.) + omtau**(2*a))**((1+g)/2.)

        return eps * np.pi * numer / denom / 2.


class ColeColeDerivative:
    name = 'Derivative CC'
    type = 'Dielectric Spectroscopy'
    params = [r'\Delta\epsilon', r'\tau', r'\alpha']
    bounds = [(0, None), (0, None), (0, 1)]

    @staticmethod
    def func(x, eps, tau, alpha):
        omtau = 2*np.pi*x * tau
        theta = np.arctan2(np.sin(np.pi*alpha/2.), omtau**(-alpha) + np.cos(np.pi*alpha/2.))

        numer = alpha * omtau**alpha * np.cos(np.pi*alpha/2. - 2*theta)
        denom = 1 + 2*omtau**alpha * np.cos(np.pi*alpha/2.) + omtau**(2*alpha)

        return eps * np.pi * numer / denom / 2.


class ColeDavidsonDerivative:
    name = 'Derivative CD'
    type = 'Dielectric Spectroscopy'
    params = [r'\Delta\epsilon', r'\tau', r'\gamma']
    bounds = [(0, None), (0, None), (0, 1)]

    @staticmethod
    def func(x, eps, tau, g):
        omtau = 2*np.pi*x * tau
        numer = g * omtau * np.sin((1+g)*np.sin(omtau))
        denom = (1 + omtau**2)**((1+g)/2.)

        return eps * np.pi * numer / denom / 2.
